library xml_layout;

import 'dart:collection';
import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:xml/xml.dart' as xml;
import 'parser.dart';
import 'exceptions.dart';
import 'types/function.dart' as function;
import 'types/proxy.dart' as proxy;

typedef ItemConstructor = dynamic Function(NodeData node, Key key);
typedef MethodConstructor = dynamic Function();
typedef _NodeTester = bool Function(NodeData);
typedef ApplyFunction = dynamic Function(String name, List args);

mixin _NodeControl {
  Map<String, GlobalKey> _keys = Map();

  GlobalKey _getKey(String id) {
    if (_keys.containsKey(id)) return _keys[id];
    return _keys[id] = GlobalKey();
  }

  GlobalKey find(String id) {
    return _keys[id];
  }

  Map<String, dynamic> get objects;
  ApplyFunction get apply;

  ItemConstructor onUnkown;
}

class _NodeInfo {
  xml.XmlNode node;
  Map<String, dynamic> extension;

  _NodeInfo(this.node, this.extension);
}

class _FlowResult {
  String tag;
  dynamic data;
  List<_NodeInfo> list;
}

abstract class _FlowControl {
  _FlowResult process(NodeData node, [_FlowResult prev]);
}

class _ForFlowControl extends _FlowControl {

  @override
  _FlowResult process(NodeData node, [_FlowResult prev]) {
    xml.XmlElement element = node.node as xml.XmlElement;
    String item = element.getAttribute("item") ?? "item";
    String index = element.getAttribute("index") ?? "index";
    xml.XmlNode fnode = element.getAttributeNode("array");
    NodeData array = fnode != null ? NodeData(fnode, node.control, node) : null;
    fnode = element.getAttributeNode("count");
    NodeData count = fnode != null ? NodeData(fnode, node.control, node) : null;

    List<_NodeInfo> res = [];
    List arr;
    if (array != null) {
      arr = array.t<Iterable>();
    } else {
      arr = [];
      dynamic len = count.t<int>();
      if (len is int && len > 0) {
        for (int i = 0; i < len; ++i) {
          arr.add(i);
        }
      }
    }

    int idx = 0;
    for (dynamic data in arr) {
      for (var child in element.children) {
        res.add(_NodeInfo(child, {item: data, index: idx++}));
      }
    }
    return _FlowResult()..list = res;
  }
}

class _IfFlowControl extends _FlowControl {

  @override
  _FlowResult process(NodeData node, [_FlowResult prev]) {
    xml.XmlElement element = node.node as xml.XmlElement;
    xml.XmlNode fnode = element.getAttributeNode("candidate");
    NodeData candidate = fnode != null ? NodeData(fnode, node.control, node) : null;

    List<_NodeInfo> res = [];
    bool candi = candidate == null ? false : candidate.boolean;
    if (candi) {
      for (var child in element.children) {
        res.add(_NodeInfo(child, null));
      }
    }
    return _FlowResult()
      ..tag = "else"
      ..data = candi
      ..list = res;
  }
}

class _ElseFlowControl extends _FlowControl {

  @override
  _FlowResult process(NodeData node, [_FlowResult prev]) {
    List<_NodeInfo> res = [];
    bool candi = true, checkCandi = prev?.data == false;
    if (checkCandi) {
      xml.XmlElement element = node.node as xml.XmlElement;
      xml.XmlNode fnode = element.getAttributeNode("candidate");
      NodeData candidate = fnode != null ? NodeData(fnode, node.control, node) : null;

      candi = candidate == null ? true : candidate.boolean;
      if (candi) {
        for (var child in element.children) {
          res.add(_NodeInfo(child, null));
        }
      }
    }
    return _FlowResult()
      ..tag = "else"
      ..data = candi || !checkCandi
      ..list = res;
  }
}

Map<String, _FlowControl> _flowControls = {
  "for": _ForFlowControl(),
  "if": _IfFlowControl(),
  "else": _ElseFlowControl(),
};

class _Range {
  int start;
  int end;
  String value;
  _Range(this.start, this.end, this.value);
}

class NodeData {
  xml.XmlNode node;
  _NodeControl control;
  Map<String, dynamic> _ext;
  NodeData _father;

  Map<String, List<NodeData>> _attributes;
  List<NodeData> _children;

  NodeData(this.node, this.control, [NodeData father]) : _father = father;

  void _setNode(String name, NodeData node) {
    name = name.toLowerCase();
    List<NodeData> list = _attributes[name];
    if (list == null) {
      list = [];
      _attributes[name] = list;
    }
    list.add(node);
  }
  
  List<NodeData> get rawChildren {
    if (_children == null) _processNode();
    return _children;
  }

  void _processChildren(xml.XmlElement element) {
    for (var child in element.children) {
      if (child is xml.XmlElement) {
        _processChildElement(child);
      }
    }
    _prevResult = null;
  }

  void _addChildrenElement(xml.XmlElement element, Map<String, dynamic> extension) {
    if (element.name.prefix == "attr") {
      List<xml.XmlElement> els = [];
      xml.XmlText text;
      for (xml.XmlNode node in element.children) {
        if (node is xml.XmlElement) {
          els.add(node);
        } else if (text == null && node is xml.XmlText) {
          text = node;
        }
      }

      if (els.length == 0) {
        _setNode(element.name.local, NodeData(text ?? xml.XmlText(""), control, this).._ext = extension);
      } else {
        els.forEach((el) {
          _setNode(element.name.local,
              NodeData(el, control, this).._ext = extension);
        });
      }
    } else if (element.name.prefix == null) {
      _children.add(NodeData(element, control, this).._ext = extension);
    } else if (element.name.prefix == "arg") {
      xml.XmlElement el;
      xml.XmlText text;
      for (xml.XmlNode node in element.children) {
        if (el == null && node is xml.XmlElement) {
          el = node;
        } else if (text == null && node is xml.XmlText) {
          text = node;
        }
      }

      _setNode(element.name.toString(),
          NodeData(el ?? text ?? xml.XmlText(""), control, this).._ext = extension);
    }
  }

  _FlowResult _prevResult;
  void _processChildElement(xml.XmlElement element, [Map<String, dynamic> extension]) {
    String tagName = element.name.toString().toLowerCase();
    if (_flowControls.containsKey(tagName)) {
      var flowControl = _flowControls[tagName];
      _prevResult = flowControl.process(NodeData(element, control, this).._ext = extension, _prevResult?.tag == tagName ? _prevResult : null);
      _prevResult.list.forEach((element) {
        if (element.node is xml.XmlElement) {
          _processChildElement(element.node, element.extension);
        }
      });
    } else {
      _prevResult = null;
      _addChildrenElement(element, extension);
    }
  }

  void _processNode() {
    if (_children == null) {
      xml.XmlElement element = node as xml.XmlElement;
      _attributes = Map();
      element.attributes.forEach((element) {
        if (element.name.prefix == null || element.name.prefix == "attr")
          _setNode(element.name.local, NodeData(element, control, this));
      });
      _children = [];
      if (node is xml.XmlElement) {
        _processChildren(node);
      }
    }
  }

  NodeData clone(Map<String, dynamic> ext) {
    NodeData one = NodeData(node, control, _father);
    one._ext = ext ?? _ext;
    one._attributes = _attributes;
    one._children = _children?.map<NodeData>((element) {
      var child = element.clone(null);
      child._father = one;
      return child;
    })?.toList();
    return one;
  }

  NodeData _firstChild(List<NodeData> nodes, [bool Function(NodeData) tester]) {
    if (nodes != null) {
      for (var node in nodes) {
        if (tester == null || tester(node)) {
          return node;
        }
      }
    }
    return null;
  }

  List<T> _convertListTo<T>(List<NodeData> nodes) {
    List<T> res = [];
    void addResult(NodeData node) {
      dynamic obj = node.t<T>();
      if (obj != null) res.add(obj);
    }

    for (var node in nodes) {
      addResult(node);
    }
    return res;
  }

  NodeData operator [](String name) {
    if (_attributes == null && node is xml.XmlElement) {
      _processNode();
    }
    if (_attributes == null) return null;
    return _firstChild(_attributes[name.toLowerCase()]);
  }

  Key _getKey() {
    NodeData id = this["id"];
    Key key;
    if (id != null) {
      key = control._getKey(id.text);
    }
    return key;
  }

  dynamic element() {
    if (isElement) {
      xml.XmlElement element = this.node as xml.XmlElement;
      _ItemInfo info =
          XmlLayout._constructors[element.name.toString().toLowerCase()];
      if (info != null) {
        return info.constructor(this, _getKey());
      }
    }
  }

  String _processText(String text) {
    const String mark = r"$";
    const String slash = r"/";

    int off = text.indexOf(mark);
    List<_Range> ranges = [];
    while (off >= 0) {
      int slashCount = 0;
      int pre = off - 1;
      while (pre > 0 && text[pre] == slash) {
        ++slashCount;
        --pre;
      }

      if (slashCount % 2 == 0) {
        if (text[off + 1] == "{") {
          Match match =
              RegExp(r"(?<=\$)\{([^\}]+)\}").matchAsPrefix(text, off + 1);
          if (match != null) {
            ranges.add(_Range(match.start - 1, match.end, match.group(1)));
          } else {}
        } else {
          Match match = RegExp(r"(?<=\$)[\w_]+").matchAsPrefix(text, off + 1);
          if (match != null) {
            ranges.add(_Range(match.start - 1, match.end, match.group(0)));
          } else {}
        }
      }
      off = text.indexOf(mark, off + 1);
    }
    ranges.reversed.forEach((element) {
      text = text.replaceRange(
          element.start, element.end, _get(element.value).toString());
    });
    return text;
  }

  String _raw;
  String get raw =>
      _raw ??
      (_raw = node is xml.XmlAttribute
          ? (node as xml.XmlAttribute).value
          : node.text.trim());
  String get text => _processText(raw);
  int get integer => int.tryParse(text);
  double get real => double.tryParse(text);
  bool get boolean {
    MethodNode obj = MethodNode.parse(raw, _get);
    return obj == null ? text == "true" : obj.execute() == true;
  }

  bool get isAttribute => node is xml.XmlAttribute;
  bool get isElement => node is xml.XmlElement;
  String get name =>
      node is xml.XmlElement ? (node as xml.XmlElement).name.local : null;

  T child<T>() {
    if (rawChildren != null) {
      T res;
      String text;
      if (_children.isEmpty && (text = this.text).isNotEmpty) {
        res = v<T>(text);
      } else {
        _firstChild(_children, (node) {
          dynamic obj = node.element();
          if (obj != null && obj is T) {
            res = obj;
            return true;
          } else {
            if (obj == null) {
              control.onUnkown?.call(node, node._getKey());
            }
            return false;
          }
        });
      }
      return res;
    }
    return null;
  }

  List<T> children<T>() {
    _processNode();

    return _children == null ? [] : _convertListTo<T>(_children);
  }

  List<T> array<T>(String name) {
    List<NodeData> attrs = _attributes[name.toLowerCase()];
    var attr = attrs?.first;
    if (attr == null) return null;
    else if (attr.isElement) {
      return _convertListTo<T>(attrs);
    } else {
      return attr.t<List<T>>();
    }
  }

  dynamic _get(String path) {
    RegExp exp = RegExp(r"^(\w+)((\[[^\]]+\])*)$");
    RegExp bExp = RegExp(r"\[([^\]]+)\]");
    List<String> arr = path.split(".");
    List segs = [];
    for (String seg in arr) {
      RegExpMatch match = exp.firstMatch(seg);
      if (match == null) return null;
      String name = match.group(1);
      String property = match.group(2);
      segs.add(name);
      if (property.length > 0) {
        var matches = bExp.allMatches(property);
        for (Match match in matches) {
          segs.add(jsonDecode(match.group(1)));
        }
      }
    }

    dynamic ret = _getPath(_ext, segs, 0);
    if (ret == null && control.objects != null) {
      ret = _getPath(control.objects, segs, 0);
    }
    if (ret == null) {
      NodeData father = _father;
      while (father != null) {
        ret = _getPath(father._ext, segs, 0);
        if (ret != null) break;
        father = father._father;
      }
    }
    return ret;
  }

  static _getPath(dynamic tar, List path, int offset) {
    if (offset >= path.length) {
      return tar;
    } else {
      dynamic seg = path[offset];
      if (tar is Map || tar is MapMixin) {
        var sub = seg is String ? tar[seg] : null;
        if (sub == null) return null;
        return _getPath(sub, path, offset + 1);
      } else if (tar is List || tar is ListMixin) {
        if (seg is int) {
          var sub = tar[seg];
          if (sub == null) return null;
          return _getPath(sub, path, offset + 1);
        } else
          return null;
      }
    }
  }

  T t<T>() {
    if (!isElement) {
      dynamic obj = MethodNode.parseString(raw, _get);
      if (obj is T) return obj;
      if (T == String) return obj.toString() as T;
    }
    switch (T) {
      case String:
        {
          return text as T;
        }
      case bool:
        {
          return boolean as T;
        }
      case int:
        {
          return integer as T;
        }
      case double:
        {
          return real as T;
        }
      default:
        {
          if (isElement) {
            dynamic obj = element();
            if (obj is T)
              return obj;
            else
              return control.onUnkown?.call(this, _getKey());
          }
          _ItemInfo info = XmlLayout._constructors[T];
          if (info == null) {
            return control.onUnkown?.call(this, _getKey());
          } else {
            dynamic obj = info.constructor(this, _getKey());
            return obj;
          }
        }
    }
  }
  T convert<T>() => t<T>();

  T s<T>(String name, [T def]) => this[name]?.t<T>() ?? def;
  T attribute<T>(String attributeName, [T defaultValue]) => s<T>(name, defaultValue);

  T v<T>(String txt, [T def]) {
    _ItemInfo info = XmlLayout._constructors[T];
    return info.constructor(NodeData(xml.XmlText(txt), control, this), null) ?? def;
  }
  T value<T>(String value, [T defaultValue]) => v<T>(value, defaultValue);

  MethodNode _arguments;
  bool _argvInit = false;
  static const String MethodPattern = r"^{0}\(([^\)]*)\)$";
  MethodNode splitMethod(String name, int count) {
    var argv = arguments;
    if (argv != null && argv.name == name && argv.length == count) {
      return argv;
    }
    return null;
  }

  MethodNode get arguments {
    if (!isElement) {
      if (!_argvInit) {
        _arguments = MethodNode.parse(text, _get);
        _argvInit = true;
      }
    }
    return _arguments;
  }

  dynamic get inlineValue {

  }

  dynamic apply(String name, List args) {
    return control.apply?.call(name, args);
  }
}

abstract class _ItemInfo {
  _ItemInfo();

  dynamic constructor(NodeData node, Key key);
}

class _ConstructorItemInfo extends _ItemInfo {
  ItemConstructor _constructor;

  _ConstructorItemInfo(this._constructor);

  @override
  constructor(NodeData node, Key key) {
    return _constructor(node, key);
  }
}

typedef InlineItemConstructor = Function(NodeData node, MethodNode method);

class _InlineItemData {
  String name;
  bool field;

  InlineItemConstructor constructor;
}

class _InlineItemInfo extends _ItemInfo {

  List<_InlineItemData> dataList = [];

  _InlineItemData _find(String name, bool field) {
    for (var data in dataList) {
      if (data.name == name && data.field == field) {
        return data;
      }
    }
    return null;
  }

  @override
  constructor(NodeData node, Key key) {
    if (node.text.contains('(')) {
      var method = MethodNode.parse(node.text, node._get);
      return _find(method.name, false)?.constructor?.call(node, method);
    } else {
      return _find(node.text, true)?.constructor?.call(node, null);
    }
  }
}

class XmlLayout extends StatefulWidget {
  static Map<dynamic, _ItemInfo> _constructors = Map();

  final xml.XmlElement element;
  final String template;
  final Map<String, dynamic> objects;
  final ApplyFunction apply;
  final ItemConstructor onUnkownElement;

  static bool _initialized = false;

  /// Constructs a XmlLayout widget
  ///
  /// [template] is the xml string to build widget.
  /// [objects] parameters will be pass to builder.
  /// [apply] handle the invoke from XmlLayout widget.
  /// [onUnkownElement] handle unkown element
  ///
  /// Example:
  ///
  /// ```dart
  /// XmlLayout(
  ///   template: "<Text>$counter</Text>",
  ///   objects: {"counter": _counter},
  /// );
  /// ```
  ///
  XmlLayout(
      {Key key, @required this.template, this.objects, this.apply, this.onUnkownElement})
      : element = null,
        super(key: key) {
    assert(template != null);
    _initialize();
  }

  /// Constructs a XmlLayout widget
  ///
  /// use a [XmlElement] to constructs a widget.
  XmlLayout.element(
      {Key key, @required this.element, this.objects, this.apply, this.onUnkownElement})
      : template = null,
        super(key: key) {
    assert(element != null);
    _initialize();
  }

  static void _initialize() {
    if (!_initialized) {
      function.register();
      proxy.register();
      _initialized = true;
    }
  }

  @override
  State<StatefulWidget> createState() => XmlLayoutState()..onUnkown = onUnkownElement;

  static Iterable<String> get registerTypes {
    Map<String, String> types = Map();
    void add(String type) {
      String lower = type.toLowerCase();
      if (!types.containsKey(lower))
        types[lower] = type;
    }

    List<String> strArr = [];
    for (var type in _constructors.keys) {
      if (type is Type) {
        add(type.toString());
      } else if (type is String) {
        strArr.add(type);
      }
    }

    for (var type in strArr) {
      var arr = type.split(".");
      add(arr[0]);
    }
    return types.values;
  }

  /// Register a constructor method, which is used to convert
  /// xml tag to a dart Object
  static void register(String name, ItemConstructor constructor) {
    _constructors[name.toLowerCase()] = _ConstructorItemInfo(constructor);
  }

  /// Shortcat to register a enum type.
  ///
  /// Example:
  ///
  /// ```dart
  /// XmlLayout.registerEnum(Brightness.values);
  /// ```
  static void registerEnum<T>(List<T> values) {
    Map<String, T> map = Map();
    values.forEach((element) {
      map[element.toString().split(".").last] = element;
    });
    _constructors[T] = _ConstructorItemInfo((node, _) {
      String name = node.text;
      return map[name];
    });
  }

  /// Register a constructor method, which is used to convert
  /// a attribute data to a Object.
  static void registerInline(Type type, String name, bool field, InlineItemConstructor constructor) {
    var item = _constructors[type];
    if (item == null) {
      _constructors[type] = item = _InlineItemInfo();
    }
    if (item is _InlineItemInfo) {
      item.dataList.add(_InlineItemData()
          ..name = name
          ..field = field
          ..constructor = constructor
      );
    } else {
      throw "Target is not a inline item";
    }
  }
}

///
/// If a element has a [id] attribute, you can use
/// [find] method to get the [GlobalKey] of built widget
/// by the id string.
///
class XmlLayoutState extends State<XmlLayout> with _NodeControl {
  NodeData _data;

  @override
  Widget build(BuildContext context) {
    if (_data == null) {
      xml.XmlElement element =
          widget.element ?? xml.XmlDocument.parse(widget.template)?.firstChild;
      if (element != null) {
        _data = NodeData(element, this);
      } else {
        throw TemplateException("Can not parse template.");
      }
    }

    dynamic tar = _data.element();
    if (tar is Widget) {
      return tar;
    } else {
      throw NotWidgetException();
    }
  }

  @override
  void didUpdateWidget(XmlLayout oldWidget) {
    super.didUpdateWidget(oldWidget);

    if (oldWidget.template != widget.template ||
        oldWidget.element != widget.element) {
      _data = null;
      _keys.clear();
    }
  }

  @override
  Map<String, dynamic> get objects => widget.objects;
  @override
  ApplyFunction get apply => widget.apply;
}

class XmlLayoutBuilder with _NodeControl {
  NodeData _data;
  Map<String, GlobalKey> _keys = Map();
  String template;
  xml.XmlElement element;
  Map<String, dynamic> _objects;
  ApplyFunction _apply;

  Widget build(BuildContext context,
      {Map<String, dynamic> objects,
      String template,
      xml.XmlElement element,
      ItemConstructor onUnkownElement,
      ApplyFunction apply}) {
    XmlLayout._initialize();
    onUnkown = onUnkownElement;
    _objects = objects;
    _apply = apply;
    if (template != this.template) {
      this.template = template;
      _data = null;
    }
    if (this.element != element) {
      this.element = element;
      _data = null;
    }
    if (_data == null) {
      xml.XmlElement el =
          element ?? xml.XmlDocument.parse(template)?.firstChild;
      if (el != null) {
        _data = NodeData(el, this);
      } else {
        throw TemplateException("Can not parse template.");
      }
    }

    dynamic tar = _data.element();
    if (tar is Widget) {
      return tar;
    } else {
      throw NotWidgetException();
    }
  }

  GlobalKey find(String id) {
    return _keys[id];
  }

  @override
  Map<String, dynamic> get objects => _objects;

  @override
  ApplyFunction get apply => _apply;
}
